#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include "hostproc.h"

struct tokbuf {
	char const *val;
	char const *cur;
	char *tokptr;
	size_t toksize;
	int reuse;
};

static void
tokbuf_init(struct tokbuf *tb, char const *val)
{
	tb->val = tb->cur = val;
	tb->tokptr = NULL;
	tb->toksize = 0;
	tb->reuse = 0;
}

static void
tokbuf_free(struct tokbuf *tb)
{
	free(tb->tokptr);
}

static char *
tokbuf_morespace(struct tokbuf *tb, size_t len)
{
	while (len > tb->toksize) {
		char *p = a2nrealloc(tb->tokptr, &tb->toksize, 1);
		if (!p)
			abend_nomem();
		tb->tokptr = p;
	}
	return tb->tokptr;
}

static int
tokbuf_eof(struct tokbuf *tb)
{
	if (tb->reuse)
		return 0;
	tb->cur += strspn(tb->cur, " \t");
	return *tb->cur == 0;
}

static char *
getquote(struct tokbuf *tb)
{
	size_t len;
	size_t nesc = 0;
	char *ret, *p;

	tb->cur++; /* Skip initial quote character */
	/* Scan the line to determine its length */
	for (len = 0; tb->cur[len] != '"'; len++) {
		if (!tb->cur[len]) {
			config_perror("unterminated quoted string near");
			return NULL;
		}
		if (tb->cur[len] == '\\') {
			nesc++;
			len++;
			if (!tb->cur[len]) {
				config_perror("unterminated quoted string near");
				return NULL;
			}
		}
	}

	ret = tokbuf_morespace(tb, len - nesc + 1);
	p = ret;
	while (*tb->cur != '"') {
		if (*tb->cur == '\\')
			tb->cur++;
		*p++ = *tb->cur++;
	}
	*p = 0;
	tb->cur++;
	
	return ret;
}
	
static char *
getstr(struct tokbuf *tb)
{
	size_t len;
	char *ret;

	if (tb->reuse) {
		tb->reuse = 0;
		return tb->tokptr;
	}
	
	len = strspn(tb->cur, " \t");
	tb->cur += len;
	if (*tb->cur == 0) {
		config_perror("expected string, but found end of line");
		return NULL;
	}
	if (*tb->cur == '"') 
		return getquote(tb);
	len = strcspn(tb->cur, " \t");

	ret = tokbuf_morespace(tb, len + 1);
	memcpy(ret, tb->cur, len);
	ret[len] = 0;
	tb->cur += len;

	return ret;
}

static int
getnum(struct tokbuf *tb, unsigned long *retval, char const *what)
{
	unsigned long n;
	char const *input;
	char *p;
	
	if (tb->reuse) {
		input = tb->tokptr;
	} else {
		tb->cur += strspn(tb->cur, " \t");
		if (*tb->cur == 0) {
			config_perror("expected number, but found end of line");
			return -1;
		}
		input = tb->cur;
	}
	n = strtoul(input, &p, 10);
	if ((n == UINT_MAX && errno == ERANGE) ||
	    (*p && !(*p == ' ' || *p == '\t'))) {
		char errbuf[128];
		snprintf(errbuf, sizeof(errbuf), "bad %s", what);
		config_perror(errbuf);
		tb->reuse = 0;
		return -1;
	}
	*retval = n;
	if (tb->reuse)
		tb->reuse = 0;
	else
		tb->cur = p;
	return 0;
}

static int
getnum_dfl(struct tokbuf *tb, unsigned long *retval, char const *what,
	   unsigned long dfl)
{
	if (tokbuf_eof(tb)) {
		*retval = dfl;
		return 0;
	}
	return getnum(tb, retval, what);
}

static inline void
putback(struct tokbuf *tb)
{
	tb->reuse = 1;
}

static void
parse_procpattern_options(struct tokbuf *tb, struct procpattern_options *opts)
{
	while (1) {
		char *p = getstr(tb);
		if (strcmp(p, "-rx") == 0)
			opts->rx = 1;
		else if (strcmp(p, "-comm") == 0)
			opts->field = pf_comm;
		else if (strcmp(p, "-exe") == 0)
			opts->field = pf_exe;
		else if (strcmp(p, "-cmdline") == 0) {
			opts->field = pf_cmdline;
			opts->rx = 1;
		} else {
			putback(tb);
			break;
		}
	}
}


int
procpattern_init(PROCPATTERN *pat, char const *str,
		 struct procpattern_options *opts)
{
	if ((pat->str = strdup(str)) == NULL)
		abend_nomem();
	pat->field = opts->field;
	if (opts->rx) {
		int ec;
		if ((pat->rx = malloc(sizeof(*pat->rx))) == NULL)
			abend_nomem();
		ec = regcomp(pat->rx, pat->str, REG_EXTENDED|REG_NOSUB);
		if (ec) {
			char errbuf[512];
			regerror(ec, pat->rx, errbuf, sizeof(errbuf));
			config_perror(errbuf);
			free(pat->rx);
			free(pat->str);
			pat->rx = NULL;
			pat->str = NULL;
			return -1;
		}
	}
	return 0;
}

void
procpattern_deinit(PROCPATTERN *pat)
{
	if (pat) {
		if (pat->rx) {
			regfree(pat->rx);
			free(pat->rx);
		}
		free(pat->str);
	}
}

int
procpattern_match(PROCESS *proc, PROCPATTERN *pat)
{
	char *str;

	switch (pat->field) {
	case pf_comm:
		str = proc->comm;
		break;
	case pf_exe:
		str = proc->exe;
		break;
	case pf_cmdline:
		str = proc->cmdline;
		break;
	default:
		abort();
	}
	if (!str)
		return 1;
	if (pat->rx)
		return regexec(pat->rx, str, 0, NULL, 0);
	return strcmp(str, pat->str);
}

PROCPATTERN_ENTRY *
procpattern_entry_create(char const *str, struct procpattern_options *opts)
{
	PROCPATTERN_ENTRY *ent;

	if ((ent = calloc(1, sizeof(*ent))) == NULL)
		abend_nomem();
	procpattern_init(&ent->pattern, str, opts);
	return ent;
}

void
procpattern_entry_free(PROCPATTERN_ENTRY *ent)
{
	procpattern_deinit(&ent->pattern);
	free(ent);
}

PROCPATTERN_ENTRY *
procpattern_list_match_first(PROCESS *proc, PROCPATTERN_LIST *patlist)
{
	PROCPATTERN_ENTRY *ent;

	STAILQ_FOREACH(ent, patlist, next) {
		if (procpattern_match(proc, &ent->pattern) == 0)
			return ent;
	}
	return NULL;
}

PROCPATTERN_ENTRY *
procpattern_list_match_next(PROCESS *proc, PROCPATTERN_ENTRY *ent)
{
	while (ent && (ent = STAILQ_NEXT(ent, next)) != NULL) {
		if (procpattern_match(proc, &ent->pattern) == 0)
			break;
	}
	return ent;
}

PROCLIM *
proclim_create(char const *str, struct procpattern_options *opts)
{
	PROCLIM *ent;

	if ((ent = calloc(1, sizeof(*ent))) == NULL)
		abend_nomem();
	if (procpattern_init(&ent->pattern, str, opts)) {
		free(ent);
		return NULL;
	}
	return ent;
}

void
proclim_free(PROCLIM *ent)
{
	procpattern_deinit(&ent->pattern);
	free(ent);
}

PROCLIM_LIST process_limits = PROCLIM_LIST_INITIALIZER(process_limits);

static void
parse_config_proc(const char *kw, char *val)
{
	char *p;
	PROCLIM *ent;
	struct tokbuf tb;
	struct procpattern_options opts = PROCPATTERN_OPTIONS_INITIALIZER;
	
	tokbuf_init(&tb, val);

	parse_procpattern_options(&tb, &opts);
	p = getstr(&tb);
	ent = proclim_create(p, &opts);
	if (ent) {
		if (getnum_dfl(&tb, &ent->max_count, "max count", 0) == 0
		    && getnum_dfl(&tb, &ent->min_count, "min count", 0) == 0) {
			if (!tokbuf_eof(&tb))
				config_pwarn("garbage after maximum count ignored");
			STAILQ_INSERT_TAIL(&process_limits,
					   ent,
					   next);
		} else
			proclim_free(ent);
	}
	tokbuf_free(&tb);
}

PROCPATTERN_LIST exclude_list = PROCPATTERN_LIST_INITIALIZER(exclude_list);

static void
parse_config_exclude(const char *kw, char *val)
{
	struct tokbuf tb;
	char *p;
	PROCPATTERN_ENTRY *ent;
	struct procpattern_options opts = PROCPATTERN_OPTIONS_INITIALIZER;
	
	tokbuf_init(&tb, val);
	parse_procpattern_options(&tb, &opts);
	p = getstr(&tb);
	ent = procpattern_entry_create(p, &opts);
	if (ent)
		STAILQ_INSERT_TAIL(&exclude_list, ent, next);
	tokbuf_free(&tb);
}

static void
parse_config_ps_interval(const char *kw, char *val)
{
	unsigned long n;
	struct tokbuf tb;
	
	tokbuf_init(&tb, val);

	if (getnum(&tb, &n, "interval"))
		return;
	if (n > UINT_MAX) {
		config_perror("interval out of allowed range");
		return;
	}
	ps_interval = n;
	tokbuf_free(&tb);
}

static void
parse_config_pidfile(const char *kw, char *val)
{
	pidfile = strdup(val);
}

struct hostproc_mib_config {
        char *token;
        char *help;
        void (*parser)(const char *token, char *line);
        void (*releaser)(void);
	void (*fini)(void);
};

static struct hostproc_mib_config config[] = {
	{ "exclude", "[-rx] [-comm] [-exe] [-cmdline] STRING",
	  parse_config_exclude, NULL },
	{ "psInterval", "NUMBER",
	  parse_config_ps_interval },
	{ "pidfile", "FILE",
	  parse_config_pidfile },
	{ "proc", "[-rx] [-comm] [-exe] [-cmdline] NAME [MAX] [MIN]",
	  parse_config_proc },
	{ NULL }
};
	 
void
hostproc_init_config(void)
{
	struct hostproc_mib_config *cp;
        for (cp = config; cp->token; cp++) {
                if (!register_app_config_handler(cp->token,
						 cp->parser,
						 cp->releaser,
						 cp->help))
                        snmp_log(LOG_ERR,"can't register %s config handler\n",
				 cp->token);
        }
}

void
hostproc_config_finalize(void)
{
	struct hostproc_mib_config *cp;
        for (cp = config; cp->token; cp++) {
		if (cp->fini)
			cp->fini();
	}
}

void
hostproc_mib_config_help(void)
{
        int i;

        for (i = 0; config[i].token; i++) {
                printf("%s %s\n", config[i].token, config[i].help);
        }
}

