/* hostproc - SNMP agent for monitoring detailed list of running processes 
   Copyright (C) 2020 Sergey Poznyakoff

   Hostproc is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Hostproc is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with hostproc. If not, see <http://www.gnu.org/licenses/>. */
#include <regex.h>
#include <sys/queue.h>

enum {
	EX_OK,
	EX_FAILURE,
	EX_USAGE
};

#define PROC_DIR "/proc"
#define UPTIME_FILE  "/proc/uptime"
#define PID_STAT_FILE "stat"
#define PID_STATM_FILE "statm"
#define PID_STATUS_FILE "status"
#define PID_CMDLINE_FILE "cmdline"
#define PID_EXE_FILE "exe"

#define STAT_SCAN_FMT "%d (%m[^)]) %c %d %d %d %d %d %u %lu %lu %lu %lu %lu %lu %ld %ld %ld %ld %ld %ld %llu %lu %ld %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %d %d %u %u %llu %lu %ld %lu %lu %lu %lu %lu %lu %lu %d"

#define STAT_SCAN_ARGS &pid,&comm,&state,&ppid,&pgrp,&session,&tty_nr,&tpgid,&flags,&minflt,&cminflt,&majflt,&cmajflt,&utime,&stime,&cutime,&cstime,&priority,&nice,&num_threads,&itrealvalue,&starttime,&vsize,&rss,&rsslim,&startcode,&endcode,&startstack,&kstkesp,&kstkeip,&signal,&blocked,&sigignore,&sigcatch,&wchan,&nswap,&cnswap,&exit_signal,&processor,&rt_priority,&policy,&delayacct_blkio_ticks,&guest_time,&cguest_time,&start_data,&end_data,&start_brk,&arg_start,&arg_end,&env_start,&env_end,&exit_code

#define STAT_SCAN_DCL				\
 int pid;\
 char *comm;\
 char state;\
 int ppid;\
 int pgrp;\
 int session;\
 int tty_nr;\
 int tpgid;\
 unsigned flags;\
 unsigned long minflt;\
 unsigned long cminflt;\
 unsigned long majflt;\
 unsigned long cmajflt;\
 unsigned long utime;\
 unsigned long stime;\
 long cutime;\
 long cstime;\
 long priority;\
 long nice;\
 long num_threads;\
 long itrealvalue;\
 unsigned long long starttime;\
 unsigned long vsize;\
 long rss;\
 unsigned long rsslim;\
 unsigned long startcode;\
 unsigned long endcode;\
 unsigned long startstack;\
 unsigned long kstkesp;\
 unsigned long kstkeip;\
 unsigned long signal;\
 unsigned long blocked;\
 unsigned long sigignore;\
 unsigned long sigcatch;\
 unsigned long wchan;\
 unsigned long nswap;\
 unsigned long cnswap;\
 int exit_signal;\
 int processor;\
 unsigned rt_priority;\
 unsigned policy;\
 unsigned long long delayacct_blkio_ticks;\
 unsigned long guest_time;\
 long cguest_time;\
 unsigned long start_data;\
 unsigned long end_data;\
 unsigned long start_brk;\
 unsigned long arg_start;\
 unsigned long arg_end;\
 unsigned long env_start;\
 unsigned long env_end;\
 int exit_code;

#define STAT_SCAN_COUNT 52

typedef struct process {
	int      pid;
	int      ppid;
	unsigned long uid;
	unsigned long gid;
	unsigned cpu_usage;
	unsigned long time;
	int      state;
	int      nice;
	int      priority;
	unsigned long vsize;
	unsigned long rss;
	char *comm;
	char *cmdline;
	char *exe;
	
	struct process *parent;
	TAILQ_ENTRY(process) link;
	struct process *next, *prev;
} PROCESS;

typedef struct {
	size_t count;
	TAILQ_HEAD(process_list, process) list;
} PROCLIST;

#define PROCLIST_INITIALIZER(h) { 0, TAILQ_HEAD_INITIALIZER(h.list) }

enum process_match_field {
	pf_comm,
	pf_exe,
	pf_cmdline
};

typedef struct process_pattern {
	enum process_match_field field;
	char *str;
	regex_t *rx;
} PROCPATTERN;

typedef struct process_pattern_entry {
	PROCPATTERN pattern;
	STAILQ_ENTRY(process_pattern_entry) next;
} PROCPATTERN_ENTRY;

typedef STAILQ_HEAD(, process_pattern_entry) PROCPATTERN_LIST;
#define PROCPATTERN_LIST_INITIALIZER(h) STAILQ_HEAD_INITIALIZER(h)

typedef struct process_limits {
	PROCPATTERN pattern;
	unsigned long min_count;
	unsigned long max_count;
	unsigned long count;
	STAILQ_ENTRY(process_limits) next;
} PROCLIM;

typedef STAILQ_HEAD(, process_limits) PROCLIM_LIST;
#define PROCLIM_LIST_INITIALIZER(h) STAILQ_HEAD_INITIALIZER(h)

extern unsigned long Hertz;
extern unsigned ps_interval;
extern char *pidfile;
extern PROCPATTERN_LIST exclude_list;
extern PROCLIM_LIST process_limits;

int proctab_foreach(int (*handler)(PROCESS *, void *), void *data);
void *hostproc_thr_ps(void *arg);
time_t agent_hostprocUpTime(void);

void hostproc_init_config(void);
void hostproc_config_finalize(void);
void hostproc_mib_config_help(void);

struct procpattern_options {
	int rx;
	int field;
};

#define PROCPATTERN_OPTIONS_INITIALIZER { .rx = 0, .field = pf_comm }

int procpattern_init(PROCPATTERN *, char const *str,
		     struct procpattern_options *opts);
void procpattern_deinit(PROCPATTERN *pat);

PROCPATTERN_ENTRY *procpattern_entry_create(char const *str,
					    struct procpattern_options *opts);
void procpattern_entry_free(PROCPATTERN_ENTRY *ent);

int procpattern_match(PROCESS *proc, PROCPATTERN *pat);
PROCPATTERN_ENTRY *procpattern_list_match_first(PROCESS *proc,
						PROCPATTERN_LIST *pat);
PROCPATTERN_ENTRY *procpattern_list_match_next(PROCESS *proc,
					       PROCPATTERN_ENTRY *ent);

void procpattern_list_append(PROCPATTERN **phead, PROCPATTERN **ptail,
			     PROCPATTERN *pat);

PROCLIM *proclim_create(char const *str, struct procpattern_options *opts);
void proclim_free(PROCLIM *ent);

void abend(char const *fmt, ...);
static inline void abend_nomem(void) { abend("out of memory"); }

void *a2nrealloc(void *p, size_t *pn, size_t s);
int proctab_scan(void);

void proctab_destroy(void);
void proctab_free(void);

int proctab_sieve(PROCPATTERN_LIST *);
void proctab_print(void);

void table_invalidate(void);
