/* hostproc - SNMP agent for monitoring detailed list of running processes 
   Copyright (C) 2020 Sergey Poznyakoff

   Hostproc is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Hostproc is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with hostproc. If not, see <http://www.gnu.org/licenses/>. */
#include <stdlib.h>
#include <pthread.h>
#include "hostproc.h"
#include "hostproc_mib.h"

enum {
	PROCESS_TABLE_VALID = 0x01,
	SUMMARY_TABLE_VALID = 0x02
};

static int table_validity_state;
static pthread_mutex_t table_validity_mutex = PTHREAD_MUTEX_INITIALIZER;

static int
table_validity_start(int what)
{
	int result;
	pthread_mutex_lock(&table_validity_mutex);
	result = table_validity_state & what;
	table_validity_state |= what;
	return result;
}

static void
table_validity_end(void)
{
	pthread_mutex_unlock(&table_validity_mutex);
}

void
table_invalidate(void)
{
	pthread_mutex_lock(&table_validity_mutex);
	table_validity_state = 0;
	pthread_mutex_unlock(&table_validity_mutex);
}

static int
processTable_load_row(PROCESS *pp, void *data)
{
	netsnmp_tdata *table = data;
	netsnmp_tdata_row *data_row;
	struct processTable_entry *ent;
	
	ent = SNMP_MALLOC_TYPEDEF(struct processTable_entry);
	if (!ent) 
		return SNMP_ERR_GENERR;

	memset(ent, 0, sizeof(*ent));
	ent->processIndex = pp->pid;
	ent->processPID = pp->pid;
	ent->processPPID = pp->ppid;
	ent->processUID = pp->uid;;
	ent->processGID = pp->gid;
	ent->processCPU = pp->cpu_usage;
	ent->processTime = pp->time;

	if ((ent->processState = malloc(2)) == NULL)
		goto err;
	ent->processState[0] = pp->state;
	ent->processState[1] = 0;	
	ent->processState_len = 1;
	
	ent->processNice = pp->nice;
	ent->processPriority = pp->priority;
	ent->processVSZ = pp->vsize;
	ent->processRSS = pp->rss;

	if ((ent->processComm = strdup(pp->comm)) == NULL)
		goto err;
	ent->processComm_len = strlen(ent->processComm);

	if ((ent->processCmdline = strdup(pp->cmdline)) == NULL)
		goto err;
	ent->processCmdline_len = strlen(ent->processCmdline);

	if (pp->exe) {
		if ((ent->processExe = strdup(pp->exe)) == NULL)
			goto err;
		ent->processExe_len = strlen(ent->processExe);
	} else {
		ent->processExe_len = ent->processComm_len + 2;
		if ((ent->processExe = malloc(ent->processExe_len + 1)) == NULL)
			goto err;
		ent->processExe[0] = '[';
		strcpy(ent->processExe + 1, ent->processComm);
		strcat(ent->processExe, "]");
	}
	data_row = netsnmp_tdata_create_row();
	if (!data_row)
		goto err;
	data_row->data = ent;
	netsnmp_tdata_row_add_index(data_row, ASN_INTEGER,
				    &ent->processIndex,
				    sizeof(ent->processIndex));
	netsnmp_tdata_add_row(table, data_row);

	return SNMP_ERR_NOERROR;

err:
	processTable_entry_free(ent);
	SNMP_FREE(ent);
	return SNMP_ERR_GENERR;
}

int
processTable_load(netsnmp_cache *cache, void *vmagic)
{
	netsnmp_tdata *table = (netsnmp_tdata *) vmagic;
	
	if (!table_validity_start(PROCESS_TABLE_VALID)) {
		int rc;
		processTable_free(cache, vmagic);
		if ((rc = proctab_foreach(processTable_load_row, table)) != SNMP_ERR_NOERROR)
			return rc;
	}
	table_validity_end();
	return 0;
}

void
processTable_entry_free(void *data)
{
	struct processTable_entry *ent = data;
	free(ent->processState);
	free(ent->processComm);
	free(ent->processCmdline);
	free(ent->processExe);
}

static void
proc_counter_init(void)
{
	PROCLIM *p;

	STAILQ_FOREACH(p, &process_limits, next)
		p->count = 0;
}

static int
proc_counter(PROCESS *pp, void *data)
{
	PROCLIM *plim = STAILQ_FIRST(&process_limits);

	while (plim) {
		if (procpattern_match(pp, &plim->pattern) == 0)
			plim->count++;
		plim = STAILQ_NEXT(plim, next);
	}
	return 0;
}

static int
proc_counter_load(netsnmp_tdata *table)
{
	size_t i;
	PROCLIM *plim;

	i = 0;
	STAILQ_FOREACH(plim, &process_limits, next) {
		struct processSummaryTable_entry *ent;
		netsnmp_tdata_row *data_row;
	
		ent = SNMP_MALLOC_TYPEDEF(struct processSummaryTable_entry);
		if (!ent) 
			return SNMP_ERR_GENERR;
		ent->processSummaryIndex = i++;
		ent->processSummaryName = plim->pattern.str;
		ent->processSummaryName_len = strlen(ent->processSummaryName);
		ent->processSummaryMin = plim->min_count;
		ent->processSummaryMax = plim->max_count;
		ent->processSummaryCount = plim->count;
		if ((plim->min_count > 0
		     && plim->count < plim->min_count)
		    || (plim->max_count > 0
			&& plim->count > plim->max_count))
			ent->processSummaryOK = 2;
		else 
			ent->processSummaryOK = 1;

		data_row = netsnmp_tdata_create_row();
		if (!data_row) {
			processSummaryTable_entry_free(ent);
			SNMP_FREE(ent);
			return SNMP_ERR_GENERR;
		}
		
		data_row->data = ent;
		netsnmp_tdata_row_add_index(data_row, ASN_INTEGER,
					    &ent->processSummaryIndex,
					    sizeof(ent->processSummaryIndex));
		netsnmp_tdata_add_row(table, data_row);
	}
	return SNMP_ERR_NOERROR;
}

int
processSummaryTable_load(netsnmp_cache *cache, void *vmagic)
{
	netsnmp_tdata *table = (netsnmp_tdata *) vmagic;
	if (!table_validity_start(SUMMARY_TABLE_VALID)) {
		proc_counter_init();
		proctab_foreach(proc_counter, NULL);
		processSummaryTable_free(cache, vmagic);
		proc_counter_load(table);
	}
	table_validity_end();	
	return SNMP_ERR_NOERROR;
}

void
processSummaryTable_entry_free(void *data)
{
	/* nothing */
}

